/* src/net/ocsp_response.c
 * ============================================================
 * OCSP Responder
 * (c) 2001 by Massimiliano Pala
 * OpenCA released software
 * ============================================================
 */
 
#include "general.h"

#include <openssl/err.h>
#include <openssl/ssl.h>

#include <openssl/txt_db.h>
#include <openssl/x509.h>

// #include <openssl/asn1.h>

#include "support.h"
#include "db.h"

#include "ocsp_response.h"


/* External General Variables */
extern int debug;
extern int verbose;

int make_ocsp_response(OCSP_RESPONSE **resp, OCSP_REQUEST *req,
		OCSPD_CONFIG *conf )
	{
	ASN1_TIME *thisupd = NULL, *nextupd = NULL;
	OCSP_CERTID *cid, *ca_id = NULL;
	OCSP_BASICRESP *bs = NULL;
	int i, id_count, ret = 1;


	id_count = OCSP_request_onereq_count(req);

	if (id_count <= 0)
		{
		*resp = OCSP_response_create(OCSP_RESPONSE_STATUS_MALFORMEDREQUEST, NULL);
		goto end;
		}

	ca_id = OCSP_cert_to_id(conf->digest, NULL, conf->cacert);

	bs = OCSP_BASICRESP_new();
	thisupd = X509_gmtime_adj(NULL, 0);
	if (conf->ndays != -1)
		nextupd = X509_gmtime_adj(NULL, conf->nmin * 60 + 
				conf->ndays * 3600 * 24 );

	/* Examine each certificate id in the request */
	for (i = 0; i < id_count; i++)
		{
		OCSP_ONEREQ *one;
		ASN1_INTEGER *serial;
		char inf;

		one = OCSP_request_onereq_get0(req, i);
		cid = OCSP_onereq_get0_id(one);

		/* Get basic request info */
		OCSP_id_get0_info(NULL, NULL, NULL, &serial, cid);

		/* Is this request about our CA? */
		if (OCSP_id_issuer_cmp(ca_id, cid)) {
			if (verbose) {
				syslog( LOG_INFO, "request for non reckognized CA",
					ASN1_INTEGER_get(serial) );
			}

			OCSP_basic_add1_status(bs, cid,
				V_OCSP_CERTSTATUS_UNKNOWN,0,NULL,thisupd,nextupd);
			continue;
		}

		if( verbose ) {
			syslog( LOG_INFO, "request for certificate serial %ld",
					ASN1_INTEGER_get( serial ) );
		}

		/* Old mechanism was using the DB lookup for certificate
		 * status, new behaviour uses the CRL lookup method */
		// inf = db_get_status(serial, conf);

		// inf = ocsp_get_status( serial );
		if(ocspd_verify_cert( conf->crl_data.crl, serial) == 1 ) {
			/* Not found in CRL, supposed valid */
			inf = DB_TYPE_VAL;
			if( verbose )
				syslog( LOG_INFO, "certificate %ld is valid",
					ASN1_INTEGER_get( serial ) );
		} else {
			/* If returned value is 0, then the cert has been
			 * found in the CRL, it is revoked */
			inf = DB_TYPE_REV;
			if( verbose )
				syslog( LOG_INFO, "certificate %ld is revoked",
					ASN1_INTEGER_get( serial ) );
		}

		if (inf == DB_TYPE_UNK) {
			if( verbose ) {
				syslog( LOG_INFO, "status unknown for %ld ",
					ASN1_INTEGER_get(serial) );
			}
			
			OCSP_basic_add1_status(bs, cid,
				V_OCSP_CERTSTATUS_UNKNOWN,0,NULL,thisupd,nextupd);

		} else if (inf == DB_TYPE_VAL) {
			if( verbose ) {
				syslog( LOG_INFO, "status VALID for %ld",
					ASN1_INTEGER_get(serial) );
			}
			
			OCSP_basic_add1_status(bs, cid,
				V_OCSP_CERTSTATUS_GOOD, 0, NULL, thisupd, nextupd);
		} else {
			ASN1_OBJECT *inst = NULL;
			ASN1_TIME *revtm = NULL;
			ASN1_GENERALIZEDTIME *invtm = NULL;
			OCSP_SINGLERESP *single;
			X509_CRL_ENTRY *entry;

			int reason = -1;
			/*
			crl_unpack_revinfo(&revtm, &reason, &inst, 
					&invtm,  db_get_revdate(serial, conf),
					conf->crl_data.crl);
			*/

			entry = ocsp_crl_get_entry( serial, conf->crl_data.crl );
			if( !entry ) {
				syslog( LOG_ERR,
					"error while retriving info from CRL");
				ret = 0;
				goto end;
			}

			/*
			crl_unpack_revinfo(&revtm, &reason, &inst, 
					&invtm,  serial, conf->crl_data.crl);
			*/

			single = OCSP_basic_add1_status(bs, cid,
						V_OCSP_CERTSTATUS_REVOKED,
						entry->reason, entry->rev_time,
						thisupd, nextupd);

			/* Actually the holdInstruction code is not supported */
			if (entry->invalidity_time)
				OCSP_SINGLERESP_add1_ext_i2d(single, 
					NID_invalidity_date, 
					entry->invalidity_time, 0, 0);
			/*
			else if (inst)
				OCSP_SINGLERESP_add1_ext_i2d(single, 
					NID_hold_instruction_code, inst, 0, 0);
			*/

			if( verbose )
				{
				if(invtm)
					syslog( LOG_INFO, "status REVOKED for %ld (inval. date)",
						ASN1_INTEGER_get(serial) );
				else if (inst)
					syslog( LOG_INFO, "status REVOKED for %ld (hold inst. code)",
						ASN1_INTEGER_get(serial) );
				else
					syslog( LOG_INFO, "status REVOKED for %ld (none)",
						ASN1_INTEGER_get(serial) );
				}

			ASN1_OBJECT_free(inst);
			ASN1_TIME_free(revtm);
			ASN1_GENERALIZEDTIME_free(invtm);
			X509_CRL_ENTRY_free( entry );
			}
		}

	OCSP_copy_nonce(bs, req);
	OCSP_basic_sign(bs, conf->ocspd_cert, conf->ocspd_pkey, 
			conf->digest, conf->other_certs,
			conf->flags);

	*resp = OCSP_response_create(OCSP_RESPONSE_STATUS_SUCCESSFUL, bs);

	end:
	ASN1_TIME_free(thisupd);
	ASN1_TIME_free(nextupd);
	OCSP_CERTID_free(ca_id);
	OCSP_BASICRESP_free(bs);
	return ret;

}

X509_CRL_ENTRY *ocsp_crl_get_entry( ASN1_INTEGER *serial, X509_CRL *crl ) {

	STACK_OF(X509_REVOKED) *rev = NULL;
	X509_REVOKED *r;
	
	int i,o;
	long loc_r;

	X509_CRL_ENTRY *entry;

	entry = (X509_CRL_ENTRY *) OPENSSL_malloc ( sizeof(X509_CRL_ENTRY));
	if( !entry ) return NULL;

	loc_r = -1;

	/* Let's have things clear :-D */
	entry->rev_time = NULL;
	entry->invalidity_time = NULL;

	rev = X509_CRL_get_REVOKED(crl);
	for(i = 0; i < sk_X509_REVOKED_num(rev); i++) {

		ASN1_ENUMERATED *asn;
		int crit, idx;

		r = sk_X509_REVOKED_value(rev, i);

		/* skip the list if it is not our entry */
		if( ASN1_INTEGER_cmp(r->serialNumber, serial))
		 	continue;

		if( ! r->extensions ) continue;

		/*
		for (o=0; o<sk_X509_EXTENSION_num(r->extensions); o++)
		{
			ASN1_OBJECT *obj;
			X509_EXTENSION *ex;
			ex=sk_X509_EXTENSION_value(r->extensions, o);
			obj=X509_EXTENSION_get_object(ex);

			syslog( LOG_INFO, "DEBUG:: extension %d NID=%d",
					o, OBJ_obj2nid( obj ));
		}
		*/


		entry->rev_time = (ASN1_TIME *) ASN1_STRING_dup(r->revocationDate);

		if( (asn = X509_REVOKED_get_ext_d2i( r, NID_crl_reason, 
				NULL, NULL )) == NULL ) {
			break;
		} else {
			loc_r = ASN1_ENUMERATED_get( asn );
			ASN1_ENUMERATED_free( asn );
		}

		entry->invalidity_time = 
			X509_REVOKED_get_ext_d2i( r, NID_invalidity_date, 
				NULL, NULL );

		break;
	}

	if( loc_r == -1 ) {
		/* error */
		return NULL;
	} else {
		entry->reason = loc_r;
	}

	/*
	switch (loc_r) {
		case -1: return NULL;
			 break;
		default:
			 entry->reason = loc_r;
	}
	*/

	return entry;

}

int X509_CRL_ENTRY_free( X509_CRL_ENTRY *e ) {

	if (e->invalidity_time)
		ASN1_GENERALIZEDTIME_free(e->invalidity_time);

	if( e->rev_time)
		ASN1_TIME_free(e->rev_time);

	return 1;
}
