/* src/net/server.c 
 * ============================================================
 * OCSP Responder
 * (c) 2001 by Massimiliano Pala
 * OpenCA released software
 * ============================================================
 */
 
#include "general.h"

#ifdef __FreeBSD__
#include <netinet/in.h>
#endif

#include <resolv.h>
#include <sys/types.h>
#include <sys/wait.h>
#include <sys/socket.h>
#include <unistd.h>

#include <openssl/err.h>
#include <openssl/ssl.h>

#include <openssl/conf.h>
#include <openssl/txt_db.h>

#include "support.h"
#include "ocsp_response.h"
#include "http_request.h"
#include "server.h"

/* External General Variables */
extern int debug;
extern int verbose;
extern OCSPD_CONFIG *ocspd_conf;

static int *child_list = NULL;
static int max_child = 1;

int start_server( char *bind_s, char *port_s, int *child_lst, int max_ch_num,
		int queue, long max_size, OCSPD_CONFIG *ocspd_conf ) {

	BIO *bio_socket, *buf_bio;

	HTTP_REQUEST *http_parsed_req = NULL;
	OCSP_REQUEST *OCSP_req = NULL;
	OCSP_RESPONSE *OCSP_resp = NULL;

	char *buf = NULL;
	char tmp_socket[1024];
	char *addr = NULL;
	int addr_len;

	int i = 0;
	long buf_len = 0;

	/* Let's copy the passed pointer of child_lst into the static
	 * variable child_list */
	child_list = child_lst;
	max_child = max_ch_num;

	/* Let's add for later SSL support */
	SSL_load_error_strings();
	OpenSSL_add_ssl_algorithms();

	/* No max size under 1024 */
	if( max_size < 1024 ) max_size = 1024;
	
	/* Alloc addr for binding to specific address:port */
	addr_len = strlen(bind_s) + strlen(port_s) + 2 ;
	if((addr = (char *) OPENSSL_malloc (addr_len)) == NULL ) {
		syslog(LOG_ERR, "Cannot alloc addr");
		exit( -5 );
	}

	/* We build the "host:port" string to be used */
	sprintf( addr, "%s:%s", bind_s, port_s );

	if( (bio_socket = init_connection( addr )) == NULL ) {
		syslog(LOG_ERR,"Can not setup socket, exit.");
		exit( -6 );
	}
	
	// Register signal handlers
	signal( SIGCHLD, handle_sigchld );
	signal( SIGHUP, handle_sighup );

	while ( bio_socket ) {
		handle_connection( &bio_socket, child_list, max_child,
				ocspd_conf );
	}

	return(0);
}

BIO *init_connection ( char *addr ) {

	BIO *bio_socket, *buf_bio;

	/* Set buffered bio */
	buf_bio = BIO_new(BIO_f_buffer());
	if (!buf_bio) {
		syslog( LOG_ERR, "Can not set bufbio %d", __LINE__ );
		return (NULL);
	}

	/* Bind to socket */
	if((bio_socket = BIO_new_accept( addr )) == NULL ) {
		syslog(LOG_ERR, "Cannot bind to %s", addr);
		return (NULL);
	}

	if (verbose) syslog(LOG_INFO, "successfully binded to %s", addr);

        BIO_set_accept_bios(bio_socket, buf_bio);
	buf_bio = NULL;
	if (BIO_do_accept(bio_socket) <= 0)
	{
		syslog( LOG_ERR, "Error setting up accept BIO");
		BIO_free_all( bio_socket );
		BIO_free( buf_bio );

		return( NULL );
	}

	return bio_socket;
}

int handle_connection( BIO **bio_socket, int *child_list, int max_child,
		OCSPD_CONFIG *ocspd_conf ) {

	int pid, i, ret = 0;
	int free_child = -1;
	int status = 0;
        BIO *curr_bio = NULL;
	BIO *out = NULL;
	BIO *pbio_socket;

	OCSP_REQUEST *ocsp_req = NULL;
	OCSP_RESPONSE *ocsp_resp = NULL;

	pbio_socket = *bio_socket;

	if (BIO_do_accept( pbio_socket ) <= 0) {
		return 0;
	}
	curr_bio = BIO_pop( pbio_socket );

	// syslog(LOG_INFO, "Handle_Connection (%d)", __LINE__ );
	while( (free_child = get_free_child( child_list, max_child )) < 0) {
		cleanup_pending_child();
	}

	pid = fork();
	if( pid < 0 ) {
		//Child Spawn error
		ret = pid;
		goto end;
	} else if ( pid > 0 ) {
		// Parent Process
		

		if( verbose ) {
			syslog( LOG_INFO, "Spawned child process [%d]", pid );
		}
		child_list[free_child] = pid;

		goto end;
	}

	signal(SIGHUP, just_die);
	signal(SIGUSR1, just_die);
	signal(SIGTERM, just_die);

	if((ocsp_req = get_ocsp_request( curr_bio )) == NULL ) {
		syslog(LOG_ERR, "Can not parse ocsp request");
		goto err;
	}

	if(( make_ocsp_response(&ocsp_resp, ocsp_req, ocspd_conf)) == 0 ) {
		syslog( LOG_ERR, "Error in generating response" );
		goto err;
	}

	send_ocsp_response( curr_bio, ocsp_resp );

	goto end;

err:
	ocsp_resp = OCSP_response_create(
			OCSP_RESPONSE_STATUS_MALFORMEDREQUEST, NULL);
	send_ocsp_response( curr_bio, ocsp_resp );

end:
	if( pid == 0 ) {
		BIO_free_all(curr_bio);
		if(out) BIO_free (out);
	}
	if( ocsp_req ) OCSP_REQUEST_free (ocsp_req );
        if( ocsp_resp) OCSP_RESPONSE_free(ocsp_resp);

	if( pid == 0 ) {
		exit(0);
	} else {
#ifdef DEBUG
		// syslog( LOG_INFO, "Child exiting correctly (%d)", ret );
		for( i = 0; i< max_child; i++ ) {
			syslog(LOG_INFO, "Slot %d, status [%d]",
					i, child_list[i] );
		}
#endif
	}

	return ret;
}


OCSP_REQUEST *get_ocsp_request( BIO *bio ) {

	int r, have_post, req_len;
	char buf[1024];
	char *header = NULL;

	OCSP_REQUEST *ocsp_req = NULL;

	req_len = 0;
	have_post = 0;
	for( ;; ) {
		r = BIO_gets(bio, buf, 1024);
		if( r <= 0 )
			return (NULL);

		if(!have_post) {
			if( strncmp( buf, "POST", 4) ) {
				syslog(LOG_ERR,"Method is not POST, rejecting");
				return (NULL);
			}
			have_post = 1;
		}

		/* Here we have seen all the header lines */
		if((buf[0] == '\n') || (buf[0] == '\r')) {
			break;
		}
	}

	ocsp_req = d2i_OCSP_REQUEST_bio(bio, NULL);
	if (!ocsp_req) {
		syslog( LOG_ERR, "Error parsing OCSP request\n");
		return(NULL);
	}

	return ocsp_req;
}

int send_ocsp_response(BIO *bio, OCSP_RESPONSE *resp) {

	char http_resp[] =
		"HTTP/1.0 200 OK\r\nContent-type: application/ocsp-response\r\n"
 		"Content-Length: %d\r\n\r\n";

	if (!bio) {
		syslog(LOG_ERR, "Error sending response, bio is NULL");
		return 0;
	}

	BIO_printf(bio, http_resp, i2d_OCSP_RESPONSE(resp, NULL));
	i2d_OCSP_RESPONSE_bio(bio, resp);
	BIO_printf(bio,"\n\n");

	BIO_flush(bio);
	return 1;
}

int get_free_child( int *child_list, int max_child ) {

	int i, ret = -1;

#ifdef DEBUG
	syslog( LOG_INFO, "get_free_child ( %d ) : max_child = %d",
			__LINE__ , max_child );
#endif
	for( i = 0; i < max_child; i++ ) {
#ifdef DEBUG
		syslog( LOG_INFO, "Checking availability of %d on %d",
				i, max_child );
#endif

		if( child_list[i] == -1 ) {
			ret = i;
			break;
		} else if( kill( child_list[i], 0 ) < 0 ) {
#ifdef DEBUG
			syslog( LOG_ERR, "Child error!!! pid %d, errno %d",
					child_list[i], errno );
			switch (errno) {
			
				case EINVAL:
					syslog( LOG_ERR, "** errno = EINVAL" );
					break;
				case ESRCH:
					syslog( LOG_ERR, 
						"** errno = ESRCH [pid %d]",
						child_list[i] );
					break;
				case EPERM:
					syslog( LOG_ERR, "** errno = EPERM" );
					break;
			}

#endif

			cleanup_pending_child();
			child_list[i] = -1;
			ret = i;
			break;
		}
	}
#ifdef DEBUG
	syslog( LOG_INFO, "Slot available for spawing child [%d]",
			ret );
#endif
	return (ret);
}

int cleanup_pending_child ( void ) {
	int i = 0, j;
	int status = 0;

#ifdef DEBUG
	syslog( LOG_INFO, "Received signal [%d]", sig);
#endif
	do {
		i = waitpid( -1, &status, WNOHANG);
		for( j = 0; j < max_child; j++ ) {
			if( child_list[j] == i )
				child_list[j] == -1;
		}
#ifdef DEBUG
		if( i > 0 ) {
			syslog( LOG_INFO, 
				"WARNING: found a pending process [ pid: %d]",
				i );
		} else {
			syslog( LOG_INFO, 
				"No Pending processes found [ pid: %d]", i );
		}
#endif
	} while ( i > 0 );
}

void handle_sigchld( int sig ) {

	cleanup_pending_child();
	return;
}

void handle_sighup( int sig ) {

	/* load crl */
	if((ocspd_conf->crl_data.crl = 
			ocspd_get_crl( ocspd_conf->crl_data.url )) == 0 ) {
		syslog ( LOG_ERR, "can not reload CRL, using the old one.");
		return;
	}

	/* load/init certificates db */
	/*
	 if(( ocspd_conf->db = 
			ocspd_load_db_file(ocspd_conf->db_file )) == NULL ) {

		syslog( LOG_ERR,"can not load/init db (%s)", 
				ocspd_conf->db_file );
		return;
	}
	*/

	syslog( LOG_INFO, "certificate db successfully reloaded (%s)",
			ocspd_conf->crl_data.url->addr );
	return;

}

void just_die ( int code ) {

#ifdef DEBUG
	syslog( LOG_INFO, "Just Die [pid %d: %d]", getpid(), getppid());
#endif
	exit(code);
};

