/* crypto/pkcs7/sign.c */
/* Copyright (C) 1995-1998 Eric Young (eay@cryptsoft.com)
 * All rights reserved.
 *
 * This package is an SSL implementation written
 * by Eric Young (eay@cryptsoft.com).
 * The implementation was written so as to conform with Netscapes SSL.
 * 
 * This library is free for commercial and non-commercial use as long as
 * the following conditions are aheared to.  The following conditions
 * apply to all code found in this distribution, be it the RC4, RSA,
 * lhash, DES, etc., code; not just the SSL code.  The SSL documentation
 * included with this distribution is covered by the same copyright terms
 * except that the holder is Tim Hudson (tjh@cryptsoft.com).
 * 
 * Copyright remains Eric Young's, and as such any Copyright notices in
 * the code are not to be removed.
 * If this package is used in a product, Eric Young should be given attribution
 * as the author of the parts of the library used.
 * This can be in the form of a textual message at program startup or
 * in documentation (online or textual) provided with the package.
 * 
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions
 * are met:
 * 1. Redistributions of source code must retain the copyright
 *    notice, this list of conditions and the following disclaimer.
 * 2. Redistributions in binary form must reproduce the above copyright
 *    notice, this list of conditions and the following disclaimer in the
 *    documentation and/or other materials provided with the distribution.
 * 3. All advertising materials mentioning features or use of this software
 *    must display the following acknowledgement:
 *    "This product includes cryptographic software written by
 *     Eric Young (eay@cryptsoft.com)"
 *    The word 'cryptographic' can be left out if the rouines from the library
 *    being used are not cryptographic related :-).
 * 4. If you include any Windows specific code (or a derivative thereof) from 
 *    the apps directory (application code) you must include an acknowledgement:
 *    "This product includes software written by Tim Hudson (tjh@cryptsoft.com)"
 * 
 * THIS SOFTWARE IS PROVIDED BY ERIC YOUNG ``AS IS'' AND
 * ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
 * IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE
 * ARE DISCLAIMED.  IN NO EVENT SHALL THE AUTHOR OR CONTRIBUTORS BE LIABLE
 * FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL
 * DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS
 * OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION)
 * HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT
 * LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY
 * OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF
 * SUCH DAMAGE.
 * 
 * The licence and distribution terms for any publically available version or
 * derivative of this code cannot be changed.  i.e. this code cannot simply be
 * copied and put under another distribution licence
 * [including the GNU Public Licence.]
 */

#include <openca/general.h>
#include <openca/sign.h>

#include <openssl/bio.h>
#include <openssl/x509.h>
#include <openssl/pem.h>
#include <openssl/err.h>

#include <openssl/asn1.h>
#include <openssl/evp.h>
#include <openssl/objects.h>
#include <openssl/bn.h>
 
#ifdef HAVE_ENGINE
#include <openssl/engine.h>
#endif

#define FORMAT_UNDEF    0
#define FORMAT_ASN1     1
#define FORMAT_TEXT     2
#define FORMAT_PEM      3
#define FORMAT_NETSCAPE 4
#define FORMAT_PKCS12   5
#define FORMAT_SMIME    6
#define FORMAT_ENGINE   7

#define NETSCAPE_CERT_HDR	"certificate"

#define APP_PASS_LEN	1024

#define PW_MIN_LENGTH 4
typedef struct pw_cb_data {
	const void *password;
	const char *prompt_info;
} PW_CB_DATA;

static char *usage[]={
"usage: sign args\n",
"\n",
" -in file        - Input data file\n",
" -out file       - File where to write to (signature)\n",
" -cert file      - Certificate file\n",
" -passin arg     - Input file pass phrase source\n",
" -keyfile file   - Signing key file\n",

#ifdef HAVE_ENGINE
" -keyform arg    - Private key format (PEM or ENGINE)\n",
" -engine e       - Use engine e, possibly a hardware device\n",
#else
" -keyform arg    - Private key format (PEM)\n",
#endif
" -othercerts file - Other certificates to be included\n",
" -key            - Password protecting the private key\n",
" -nd             - Non detached signature (data together with signature)\n",
" -no_sign        - Generate a PKCS#7 structure without signature\n",
" -verbose        - Verbose printing\n",
"\n",
NULL
};

char *app_get_pass(BIO *err, char *arg, int keepbio);
int app_passwd(BIO *err, char *arg1, char *arg2, char **pass1, char **pass2);
int str2fmt(char *s);

int main(int argc, char *argv[])
{
	/* PARAMETERS CHECKING */
	int badops=0;
	int verbose=0;
	char *passargin = NULL;
	char *infile=NULL;
	char *outfile=NULL;
	char *certfile=NULL;
	char *othercertsfile=NULL;
	char *keyfile=NULL;
	char *key=NULL;
	int keyform=FORMAT_PEM;
	BIO *bio_err=NULL;
	BIO *STDout=NULL;
	BIO *Sout=NULL;
	char **pp = NULL ;

	/* PKCS7 USED VARIABLES */
	X509 *x509;
	STACK_OF(X509) *rother = NULL;
	EVP_PKEY *pkey = NULL ;
	PKCS7 *p7 = NULL ;
	PKCS7_SIGNER_INFO *si = NULL ;
	BIO *certf = NULL , *keyf = NULL ;
	BIO *data = NULL ,*p7bio = NULL ;
	char buf[1024*4];
	int i;
	int nodetach=0;
	int is_signed=1;

	/* ADD ENGINE SUPPORT */
#ifdef HAVE_ENGINE
	char *engine=NULL;
	ENGINE *e = NULL;
#endif


	/* ADD ALL CHIPHERS */
	OpenSSL_add_all_ciphers();

	/* ADD DIGEST CYPHERS */
	OpenSSL_add_all_digests();

        if ((bio_err=BIO_new(BIO_s_file())) != NULL)
        	BIO_set_fp(bio_err,stderr,BIO_NOCLOSE|BIO_FP_TEXT);

        STDout=BIO_new_fp(stdout,BIO_NOCLOSE);

	/* ARGUMENTS PARSING */
        argc--;
        argv++;
        while (argc >= 1)
                {
                if      (strcmp(*argv,"-in") == 0)
                        {
                        if (--argc < 1) goto bad;
                        infile= *(++argv);
                        }
		else if (strcmp(*argv,"-out") == 0)
                        {
                        if (--argc < 1) goto bad;
                        outfile= *(++argv);
                        }
		else if (strcmp(*argv,"-cert") == 0)
                        {
                        if (--argc < 1) goto bad;
                        certfile= *(++argv);
                        }
		else if (strcmp(*argv,"-othercerts") == 0)
                        {
                        if (--argc < 1) goto bad;
                        othercertsfile= *(++argv);
                        }
		else if (strcmp(*argv,"-keyfile") == 0)
                        {
                        if (--argc < 1) goto bad;
                        keyfile= *(++argv);
                        }
		else if (strcmp(*argv,"-keyform") == 0)
			{
			if (--argc < 1) goto bad;
			keyform=str2fmt(*(++argv));
			}
		else if (strcmp(*argv,"-key") == 0)
                        {
                        if (--argc < 1) goto bad;
                        key= *(++argv);
                        }
		else if (strcmp(*argv,"-passin") == 0)
			{
			if (--argc < 1) goto bad;
			passargin= *(++argv);
			}
                else if (strcmp(*argv,"-no_sign") == 0)
                        is_signed=0;
                else if (strcmp(*argv,"-verbose") == 0)
                        verbose=1;
                else if (strcmp(*argv,"-nd") == 0)
                        nodetach=1;
		else
                        {
bad:
                        BIO_printf(bio_err,"unknown option %s\n",*argv);
                        badops=1;
                        break;
                        }
                argc--;
                argv++;
                }

	if( !keyfile && is_signed )
		badops=1;

	if( (!certfile) || (!infile && is_signed) ) 
		badops=1;

	if (badops)
                {
                for (pp=usage; (*pp != NULL); pp++)
                        BIO_printf(bio_err,*pp);
                	exit(1);
                }


	data=BIO_new(BIO_s_file());

	if( infile == NULL ) {
		BIO_set_fp(data,stdin,BIO_NOCLOSE);
	} else {
		if (!BIO_read_filename(data,infile))
			goto err;
	}

	ERR_load_crypto_strings();

	/* ENGINE support added */
#ifdef HAVE_ENGINE
	if (engine != NULL)
		{
		if((e = ENGINE_by_id(engine)) == NULL)
			{
			BIO_printf(bio_err,"invalid engine \"%s\"\n",
				engine);
			goto err;
			}
		if(!ENGINE_set_default(e, ENGINE_METHOD_ALL))
			{
			BIO_printf(bio_err,"can't use that engine\n");
			goto err;
			}
		BIO_printf(bio_err,"engine \"%s\" set.\n", engine);
		ENGINE_free(e);
		}
#endif
	if( is_signed ) {
		/* Passin support */
		if(!key && !app_passwd(bio_err, passargin, NULL, &key, NULL)) {
			BIO_printf(bio_err, "Error getting passwords\n");
			goto err;
		}

		/* Reading Private key file */
		if( verbose )
        		BIO_printf(STDout,"Reading Private Key file.\n");
		if ((keyf=BIO_new_file( keyfile, "r")) == NULL)
			goto err;
		if ( keyform == FORMAT_ENGINE)
		{
#ifdef HAVE_ENGINE
			if(!e) {
				BIO_printf( bio_err,"no engine loaded!\n");
				goto err;
			}
			BIO_printf(bio_err,"No real engine support, yet.\n");
			// pkey = ENGINE_load_private_key( e, keyfile, key );
#else
			BIO_printf(bio_err,"No engine support compiled.\n");
			goto err;
#endif
		} else if ( keyform == FORMAT_PEM ) {
			pkey=PEM_read_bio_PrivateKey(keyf,NULL,NULL,key);
			if (key) memset(key,0,strlen(key));
        	} else {
			BIO_printf(bio_err,"bad input format specified for key file\n");
			goto err;
		};
		if( keyf ) BIO_free( keyf );

		if( pkey == NULL ) {
			BIO_printf(bio_err,"Error loading private key\n");
			goto err;
		};
	}


	/* Read Certificate file and certficate data */
	if( verbose )
        	BIO_printf(STDout,"Reading Certificate file.\n");
	if ((certf=BIO_new_file( certfile, "r")) == NULL)
		goto err;
	if ((x509=PEM_read_bio_X509(certf,NULL,NULL,NULL)) == NULL)
		goto err;
	BIO_free( certf );

	p7=PKCS7_new();

	if( is_signed == 1 ) {
		PKCS7_set_type(p7,NID_pkcs7_signed);
	 
		si=PKCS7_add_signature(p7,x509,pkey,EVP_sha1());
		if (si == NULL) goto err;

		/* Get signing time automatically added */
		PKCS7_add_signed_attribute(si, NID_pkcs9_contentType,
				V_ASN1_OBJECT, OBJ_nid2obj(NID_pkcs7_data));

	} else {
		PKCS7_set_type(p7,NID_pkcs7_signed);
		// PKCS7_set_type(p7,NID_pkcs7_enveloped);
	}

	/* we may want to add more certificates to the structure, for example
	 * in a scep response */
	if( othercertsfile ) {
		X509 *tmp_cert = NULL;
		char name[1024];

		rother = load_certs(bio_err, othercertsfile, FORMAT_PEM,
				NULL, "other certificates");
		if (!rother) goto err;

		while( sk_X509_num( rother ) > 0 ) {
			tmp_cert = sk_X509_pop(rother);

			PKCS7_add_certificate(p7,tmp_cert);
			if( verbose ) {
				X509_NAME_oneline(X509_get_subject_name(
						tmp_cert ), name, 1024);
				BIO_printf( bio_err, "Addedd certificate %s\n",
						name);
			}
		}

	} else {
		/* we may want to add more */
		PKCS7_add_certificate(p7,x509);
	}


	/* Set the content of the signed to 'data' */
	PKCS7_content_new(p7,NID_pkcs7_data);

	if (!nodetach)
		PKCS7_set_detached(p7,1);

	if ((p7bio=PKCS7_dataInit(p7,NULL)) == NULL) goto err;

	if (verbose)
        	BIO_printf(STDout,"Reading Data to be signed.\n");
	for (;;)
	{
		i=BIO_read(data,buf,sizeof(buf));
		if (i <= 0) break;
		BIO_write(p7bio,buf,i);
	}

	if (!PKCS7_dataFinal(p7,p7bio)) goto err;
	BIO_free(p7bio);

	Sout=BIO_new(BIO_s_file());
	if (outfile != NULL)
	{
        	if (BIO_write_filename(Sout,outfile) <= 0)
                {
			BIO_printf(bio_err,"Error writing file %s\n", outfile);
                	perror(outfile);
                        goto err;
                }
        } else {
        	BIO_set_fp(Sout,stdout,BIO_NOCLOSE|BIO_FP_TEXT);
	}

	PEM_write_bio_PKCS7(Sout,p7);
	PKCS7_free(p7);

	exit(0);

err:
	ERR_print_errors_fp(stderr);
	exit(1);
	}


int str2fmt(char *s) {
	if 	((*s == 'D') || (*s == 'd'))
		return(FORMAT_ASN1);
	else if ((*s == 'T') || (*s == 't'))
		return(FORMAT_TEXT);
	else if ((*s == 'P') || (*s == 'p'))
		return(FORMAT_PEM);
	else if ((*s == 'N') || (*s == 'n'))
		return(FORMAT_NETSCAPE);
	else if ((*s == 'S') || (*s == 's'))
		return(FORMAT_SMIME);
	else if ((*s == '1')
		|| (strcmp(s,"PKCS12") == 0) || (strcmp(s,"pkcs12") == 0)
		|| (strcmp(s,"P12") == 0) || (strcmp(s,"p12") == 0))
		return(FORMAT_PKCS12);
	else if ((*s == 'E') || (*s == 'e'))
		return(FORMAT_ENGINE);
	else
		return(FORMAT_UNDEF);
	}



int app_passwd(BIO *err, char *arg1, char *arg2, char **pass1, char **pass2)
{
	int same;
	if(!arg2 || !arg1 || strcmp(arg1, arg2)) same = 0;
	else same = 1;
	if(arg1) {
		*pass1 = app_get_pass(err, arg1, same);
		if(!*pass1) return 0;
	} else if(pass1) *pass1 = NULL;
	if(arg2) {
		*pass2 = app_get_pass(err, arg2, same ? 2 : 0);
		if(!*pass2) return 0;
	} else if(pass2) *pass2 = NULL;
	return 1;
}

char *app_get_pass(BIO *err, char *arg, int keepbio) {
	char *tmp, tpass[APP_PASS_LEN];
	static BIO *pwdbio = NULL;
	int i;
	if(!strncmp(arg, "pass:", 5)) return BUF_strdup(arg + 5);
	if(!strncmp(arg, "env:", 4)) {
		tmp = getenv(arg + 4);
		if(!tmp) {
			BIO_printf(err, "Can't read environment variable %s\n", arg + 4);
			return NULL;
		}
		return BUF_strdup(tmp);
	}
	if(!keepbio || !pwdbio) {
		if(!strncmp(arg, "file:", 5)) {
			pwdbio = BIO_new_file(arg + 5, "r");
			if(!pwdbio) {
				BIO_printf(err, "Can't open file %s\n", arg + 5);
				return NULL;
			}
		} else if(!strncmp(arg, "fd:", 3)) {
			BIO *btmp;
			i = atoi(arg + 3);
			if(i >= 0) pwdbio = BIO_new_fd(i, BIO_NOCLOSE);
			if((i < 0) || !pwdbio) {
				BIO_printf(err, "Can't access file descriptor %s\n", arg + 3);
				return NULL;
			}
			/* Can't do BIO_gets on an fd BIO so add a buffering BIO */
			btmp = BIO_new(BIO_f_buffer());
			pwdbio = BIO_push(btmp, pwdbio);
		} else if(!strcmp(arg, "stdin")) {
			pwdbio = BIO_new_fp(stdin, BIO_NOCLOSE);
			if(!pwdbio) {
				BIO_printf(err, "Can't open BIO for stdin\n");
				return NULL;
			}
		} else {
			BIO_printf(err, "Invalid password argument \"%s\"\n", arg);
			return NULL;
		}
	}
	i = BIO_gets(pwdbio, tpass, APP_PASS_LEN);
	if(keepbio != 1) {
		BIO_free_all(pwdbio);
		pwdbio = NULL;
	}
	if(i <= 0) {
		BIO_printf(err, "Error reading password from BIO\n");
		return NULL;
	}
	tmp = strchr(tpass, '\n');
	if(tmp) *tmp = 0;
	return BUF_strdup(tpass);
}

STACK_OF(X509) *load_certs(BIO *err, const char *file, int format,
	const char *pass, const char *cert_descrip)
	{
	BIO *certs;
	int i;
	STACK_OF(X509) *othercerts = NULL;
	STACK_OF(X509_INFO) *allcerts = NULL;
	X509_INFO *xi;
	PW_CB_DATA cb_data;

	cb_data.password = pass;
	cb_data.prompt_info = file;

	if((certs = BIO_new(BIO_s_file())) == NULL)
		{
		ERR_print_errors(err);
		goto end;
		}

	if (file == NULL)
		BIO_set_fp(certs,stdin,BIO_NOCLOSE);
	else
		{
		if (BIO_read_filename(certs,file) <= 0)
			{
			BIO_printf(err, "Error opening %s %s\n",
				cert_descrip, file);
			ERR_print_errors(err);
			goto end;
			}
		}

	if      (format == FORMAT_PEM)
		{
		othercerts = sk_X509_new_null();
		if(!othercerts)
			{
			sk_X509_free(othercerts);
			othercerts = NULL;
			goto end;
			}
		// allcerts = PEM_X509_INFO_read_bio(certs, NULL,
		// 		(pem_password_cb *)password_callback, &cb_data);
		allcerts = PEM_X509_INFO_read_bio(certs, NULL,
				NULL, &cb_data);
		for(i = 0; i < sk_X509_INFO_num(allcerts); i++)
			{
			xi = sk_X509_INFO_value (allcerts, i);
			if (xi->x509)
				{
				sk_X509_push(othercerts, xi->x509);
				xi->x509 = NULL;
				}
			}
		goto end;
		}
	else	{
		BIO_printf(err,"bad input format specified for %s\n",
			cert_descrip);
		goto end;
		}
end:
	if (othercerts == NULL)
		{
		BIO_printf(err,"unable to load certificates\n");
		ERR_print_errors(err);
		}
	if (allcerts) sk_X509_INFO_pop_free(allcerts, X509_INFO_free);
	if (certs != NULL) BIO_free(certs);
	return(othercerts);
	}

